#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.16 22:00:00                  #
# ================================================== #

import sys

from PySide6.QtGui import QScreen
from PySide6.QtCore import QTimer, Signal, Slot
from PySide6.QtWidgets import (QApplication, QMainWindow)
from qt_material import QtStyleTools

from .config import Config
from .ui.main import UI
from .controller.main import Controller
from .debugger import Debug
from .settings import Settings
from .info import Info
from .gpt import Gpt
from .image import Image
from .utils import get_init_value

from .plugin.self_loop.plugin import Plugin as SelfLoopPlugin
from .plugin.real_time.plugin import Plugin as RealTimePlugin
from .plugin.web_search.plugin import Plugin as WebSearchPlugin
from .plugin.audio_azure.plugin import Plugin as AudioAzurePlugin


class MainWindow(QMainWindow, QtStyleTools):
    statusChanged = Signal(str)

    def __init__(self):
        """App main window"""
        super().__init__()
        self.timer = None
        self.github = get_init_value("__github__")
        self.website = get_init_value("__website__")
        self.version = get_init_value("__version__")
        self.build = get_init_value("__build__")
        self.author = get_init_value("__author__")
        self.email = get_init_value("__email__")
        self.data = {}

        # setup config
        self.config = Config()
        self.config.init()

        # app controller
        self.controller = Controller(self)
        self.debugger = Debug(self)
        self.info = Info(self)
        self.settings = Settings(self)

        # handle config migration if needed
        self.controller.launcher.migrate_version()

        # setup GPT and DALL-E
        self.gpt = Gpt(self.config)
        self.images = Image(self.config)

        # setup UI
        self.ui = UI(self)
        self.ui.setup()

        self.setWindowTitle('PYGPT.net v{} | build {}'.format(self.version, self.build))

        # setup signals
        self.statusChanged.connect(self.update_status)

    def log(self, data):
        """
        Logs data to console

        :param text: text to log
        """
        self.controller.debug.log(data)

    def set_theme(self, theme='dark_teal.xml'):
        """
        Updates theme

        :param theme: theme name
        """
        label = "#ffffff"
        inverse = False
        if theme.startswith('light_'):
            label = "#000000"
            inverse = True
        extra = {
            'density_scale': '-2',
            'pyside6': True,
            'QLineEdit': {
                'color': label,
            },
        }
        self.setStyleSheet(self.controller.theme.get_style('line_edit'))  # fix for line edit
        self.apply_stylesheet(self, theme, invert_secondary=inverse, extra=extra)

    def add_plugin(self, plugin):
        """
        Adds plugin

        :param plugin: plugin instance
        """
        plugin.attach(self)
        self.controller.plugins.register(plugin)

    def setup_plugins(self):
        self.controller.plugins.setup()

    def setup(self):
        """Setups app"""
        self.controller.setup()
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(30)

    def update(self):
        """Called on update"""
        self.debugger.update()

    def set_status(self, text):
        """
        Updates status text

        :param text: status text
        """
        self.data['status'].setText(str(text))

    @Slot(str)
    def update_status(self, text):
        self.set_status(text)

    def closeEvent(self, event):
        """
        Handles close event

        :param event: close event
        """
        print("Closing...")
        print("Saving config...")
        self.timer.stop()
        self.config.save()
        print("Saving presets...")
        self.config.save_presets()
        print("Exiting...")
        event.accept()  # let the window close


class Launcher:
    def __init__(self):
        """Launcher"""
        self.app = None
        self.window = None

    def init(self):
        """Initializes app"""
        self.app = QApplication(sys.argv)
        self.window = MainWindow()
        self.app.aboutToQuit.connect(self.app.quit)

    def add_plugin(self, plugin=None):
        """
        Registers plugin

        :param plugin: plugin instance
        """
        self.window.add_plugin(plugin)
        self.window.setup_plugins()

    def run(self):
        """Runs app"""
        self.window.setup()
        available_geometry = self.window.screen().availableGeometry()
        pos = QScreen.availableGeometry(QApplication.primaryScreen()).topLeft()
        self.window.resize(available_geometry.width() - 50, available_geometry.height() - 50)
        # self.window.showMaximized()
        self.window.show()
        self.window.move(pos)

        try:
            sys.exit(self.app.exec())
        except SystemExit:
            print("Closing...")


def run():
    """Runs app"""
    # initialize app
    launcher = Launcher()
    launcher.init()

    # add plugins
    launcher.add_plugin(SelfLoopPlugin())
    launcher.add_plugin(RealTimePlugin())
    launcher.add_plugin(WebSearchPlugin())
    launcher.add_plugin(AudioAzurePlugin())

    # run app
    launcher.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QLabel, QHBoxLayout

from ..utils import trans


class Status:
    def __init__(self, window=None):
        """
        Input UI

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """
        Setups status

        :return: QHBoxLayout
        """
        self.window.data['status'] = QLabel(trans('status.started'))
        layout = QHBoxLayout()
        layout.addWidget(self.window.data['status'])
        return layout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QApplication

from ..context import ContextItem
from ..utils import trans


class Input:
    def __init__(self, window=None):
        """
        Input controller

        :param window: main window
        """
        self.window = window

    def setup(self):
        """Sets up input"""
        if self.window.config.data['send_clear']:
            self.window.data['input.send_clear'].setChecked(True)
        else:
            self.window.data['input.send_clear'].setChecked(False)

        if self.window.config.data['send_shift_enter']:
            self.window.data['input.send_shift_enter'].setChecked(True)
            self.window.data['input.send_enter'].setChecked(False)
        else:
            self.window.data['input.send_enter'].setChecked(True)
            self.window.data['input.send_shift_enter'].setChecked(False)

        self.window.data['input'].setFocus()

    def toggle_send_clear(self, value):
        """
        Toggles send clear

        :param value: value of the checkbox
        """
        self.window.config.data['send_clear'] = value

    def toggle_send_shift(self, value):
        """
        Toggles send with shift

        :param value: value of the checkbox
        """
        self.window.config.data['send_shift_enter'] = value

    def send_text(self, text):
        """
        Sends text to GPT

        :param text: text to send
        """
        self.window.statusChanged.emit(trans('status.sending'))

        # prepare names
        self.window.log("User name: {}".format(self.window.config.data['user_name']))  # log
        self.window.log("AI name: {}".format(self.window.config.data['ai_name']))  # log

        user_name = self.window.controller.plugins.apply('user.name', self.window.config.data['user_name'])
        ai_name = self.window.controller.plugins.apply('ai.name', self.window.config.data['ai_name'])

        self.window.log("User name [after plugin: user_name]: {}".format(self.window.config.data['user_name']))  # log
        self.window.log("AI name [after plugin: ai_name]: {}".format(self.window.config.data['ai_name']))  # log

        # create ctx item
        ctx = ContextItem()
        ctx.set_input(text, user_name)
        ctx.set_output(None, ai_name)

        self.window.log("Context: input: {}".format(ctx.dump()))  # log

        # apply plugins
        ctx = self.window.controller.plugins.apply('ctx.before', ctx)

        # logger
        self.window.log("Context: input [after plugin: ctx.before]: {}".format(ctx.dump()))  # log
        self.window.log("System: {}".format(self.window.gpt.system_prompt))  # log

        # apply cfg
        self.window.gpt.user_name = ctx.input_name
        self.window.gpt.ai_name = ctx.output_name
        self.window.gpt.system_prompt = self.window.controller.plugins.apply('system.prompt',
                                                                             self.window.config.data['prompt'])

        self.window.log("System [after plugin: system.prompt]: {}".format(self.window.gpt.system_prompt))  # log
        self.window.log("User name: {}".format(self.window.gpt.user_name))  # log
        self.window.log("AI name: {}".format(self.window.gpt.ai_name))  # log
        self.window.log("Appending input to chat window...")  # log

        self.window.controller.output.append_input(ctx)

        # call GPT
        try:
            try:
                self.window.log("Calling OpenAI API...")  # log
                ctx = self.window.gpt.call(text, ctx)
                self.window.log("Context: output: {}".format(ctx.dump()))  # log
            except Exception as e:
                self.window.log("GPT output error: {}".format(e))  # log
                print("Error in send text (GPT call): " + str(e))
                self.window.ui.dialogs.alert(str(e))
                self.window.set_status(trans('status.error'))

            ctx = self.window.controller.plugins.apply('ctx.after', ctx)  # apply plugins
            self.window.log("Context: output [after plugin: ctx.after]: {}".format(ctx.dump()))  # log

            self.window.log("Appending output to chat window...")  # log
            self.window.controller.output.append_output(ctx)
            self.window.gpt.context.store()
            self.window.set_status(
                trans('status.tokens') + ": {} + {} = {}".format(ctx.input_tokens, ctx.output_tokens, ctx.total_tokens))
        except Exception as e:
            self.window.log("Output error: {}".format(e))  # log
            print("Error in send text: " + str(e))
            self.window.ui.dialogs.alert(str(e))
            self.window.set_status(trans('status.error'))

        return ctx

    def user_send(self, text=None):
        """Sends real user input"""
        text = self.window.controller.plugins.apply('user.send', text)
        self.send(text)

    def send(self, text=None):
        """
        Sends input text to API
        """
        self.window.statusChanged.emit(trans('status.sending'))

        ctx = None
        if text is None:
            text = self.window.data['input'].toPlainText().strip()

        self.window.log("Input text: {}".format(text))  # log
        text = self.window.controller.plugins.apply('input.before', text)

        self.window.log("Input text [after plugin: input.before]: {}".format(text))  # log

        if len(text) > 0:
            if self.window.config.data['send_clear']:
                self.window.data['input'].clear()

            # check API key
            if self.window.config.data['api_key'] is None or self.window.config.data['api_key'] == '':
                self.window.controller.launcher.show_api_monit()
                self.window.controller.ui.update()
                self.window.statusChanged.emit("")
                return

            # init api key if defined later
            self.window.gpt.init()
            self.window.images.init()

            # prepare context
            if len(self.window.gpt.context.contexts) == 0:
                self.window.gpt.context.new()
                self.window.controller.context.update()
                self.window.log("New context created...")  # log

            # process events to update UI
            QApplication.processEvents()

            # send to API
            if self.window.config.data['mode'] == 'img':
                ctx = self.window.controller.image.send_text(text)
            else:
                ctx = self.window.controller.input.send_text(text)
        else:
            self.window.statusChanged.emit("")

        self.window.log("Context: output: {}".format(ctx.dump()))  # log
        ctx = self.window.controller.plugins.apply('ctx.end', ctx)  # apply plugins
        self.window.log("Context: output [after plugin: ctx.end]: {}".format(ctx.dump()))  # log
        self.window.controller.ui.update()

    def append(self, text):
        """
        Appends text to input

        :param text: text to append
        """
        cur = self.window.data['input'].textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(text) + "\n"
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertBlock()
        self.window.data['input'].setTextCursor(cur)  # Update visible cursor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.16 22:00:00                  #
# ================================================== #

import json
import ssl
import time

import wikipedia
import re
from bs4 import BeautifulSoup
from urllib.request import Request, urlopen
from urllib.parse import quote


class WebSearch:
    def __init__(self, plugin=None):
        """
        Web search

        :param plugin: plugin
        """
        self.plugin = plugin

    def google_search(self, q, num):
        """
        Google search

        :param q: query string
        :param num: number of results
        :return: list of URLs
        """
        key = self.plugin.options["google_api_key"]['value']
        cx = self.plugin.options["google_api_cx"]['value']

        if key is None or cx is None or key == "" or cx == "":
            err = "Google API key or CX is not set. Please set them in plugin settings."
            print(err)
            self.plugin.window.ui.dialogs.alert(err)
            return []

        urls = []
        try:
            url = 'https://www.googleapis.com/customsearch/v1'
            url += '?key=' + str(key)
            url += '&cx=' + str(cx)
            url += '&num=' + str(num)
            url += '&sort=date-sdate:d:s'
            url += '&fields=items(link)'
            url += '&q=' + quote(q)

            self.plugin.window.log("Plugin: web_search:google_search: calling API: {}".format(url))  # log
            data = urlopen(url).read()
            res = json.loads(data)
            self.plugin.window.log("Plugin: web_search:google_search: received response: {}".format(res))  # log
            if 'items' not in res:
                return []
            for item in res['items']:
                urls.append(item['link'])
            self.plugin.window.log("Plugin: web_search:google_search [urls]: {}".format(urls))  # log
            return urls
        except Exception as e:
            self.plugin.window.log("Plugin: web_search:google_search: error: {}".format(e))  # log
            self.plugin.window.ui.dialogs.alert("Google Search Error: " + str(e))
            print("Error in google_search: " + str(e))
        return []

    def query_wiki(self, string):
        """
        Query wikipedia for a string and return the text content

        :param string: string to query
        :return: text content
        """
        try:
            # try search page
            wiki = wikipedia.page(string)
            text = wiki.content
            text = re.sub(r'==.*?==+', '', text)
            text = text.replace('\n', '')
            self.plugin.window.log("Plugin: web_search:query_wiki [content]: {}".format(text))  # log
            return text
        except Exception as ex:
            # try summary if page not found
            try:
                text = wikipedia.summary(string)
                text = re.sub(r'==.*?==+', '', text)
                text = text.replace('\n', '')
                self.plugin.window.log("Plugin: web_search:query_wiki [content]: {}".format(text))  # log
                return text
            except Exception as e:
                self.plugin.window.log("Plugin: web_search:query_wiki [error]: {}".format(e))  # log
                print("Error in query_wiki: " + str(e))

            self.plugin.window.log("Plugin: web_search:query_wiki [error]: {}".format(ex))  # log
            print("Error in query_wiki: " + str(ex))

    def query_web(self, string):
        """
        Query the web for a string and return the text content of the pages

        :param string: string to query
        :return: list of text content
        """
        self.plugin.window.log("Plugin: web_search:query_web [string]: {}".format(string))  # log
        pages = []
        if self.plugin.options["use_google"]['value']:
            urls = self.google_search(string, self.plugin.options["num_pages"]['value'])
            for url in urls:
                # if Wikipedia is enabled and URL is from Wikipedia, use Wikipedia
                if self.plugin.options["use_wikipedia"]['value'] and "wikipedia.org/" in url:
                    self.plugin.window.log("Plugin: web_search:query_web: using Wikipedia...")  # log
                    content = self.query_wiki(string)

                    # if Wiki content is empty, use Google instead
                    if content is None or content == "":
                        self.plugin.window.log("Plugin: web_search:query_web: nothing in Wiki, using Google...")  # log
                        content = self.query_url(url)
                    pages.append(content)
                else:
                    # use Google
                    self.plugin.window.log("Plugin: web_search:query_web: using Google...")  # log
                    content = self.query_url(url)
                    pages.append(content)
                self.plugin.window.log("Plugin: web_search:query_web [crawled content]: {}".format(content))  # log

                if content is not None:
                    self.plugin.window.log(
                        "Plugin: web_search:query_web [content length]: {}".format(len(content)))  # log

        # if Google is disabled, use Wikipedia
        if len(pages) == 0:
            if self.plugin.options["use_wikipedia"]['value']:
                content = self.query_wiki(string)
                self.plugin.window.log("Plugin: web_search:query_web: nothing found, using Wikipedia...")  # log
                pages.append(content)
        return pages

    def query_url(self, url):
        """
        Query a URL and return the text content

        :param url: URL to query
        :return: text content
        """
        self.plugin.window.log("Plugin: web_search:query_url: crawling URL: {}".format(url))  # log
        text = ''
        try:
            req = Request(
                url=url,
                headers={'User-Agent': 'Mozilla/5.0'}
            )
            if self.plugin.options['disable_ssl']['value']:
                context = ssl.create_default_context()
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                html = urlopen(req, context=context).read().decode("utf-8")
            else:
                html = urlopen(req).read().decode("utf-8")
            soup = BeautifulSoup(html, "html.parser")
            for paragraph in soup.find_all('p'):
                text += paragraph.text
            text = text.replace("\n", " ").replace("\t", " ")
            self.plugin.window.log("Plugin: web_search:query_url: received text: {}".format(text))  # log
            return text
        except Exception as e:
            self.plugin.window.log("Plugin: web_search:query_url: error querying: {}".format(url))  # log
            print("Error in query_web: " + str(e))

    def to_chunks(self, text, chunk_size):
        """
        Split text into chunks

        :param text: text to split
        :param chunk_size: chunk size
        :return: list of chunks
        """
        if text is None or text == "":
            return []
        return [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]

    def get_summarized_text(self, chunks, question):
        """
        Get summarized text from chunks

        :param chunks: chunks of text
        :param question: question to summarize
        :return: summarized text
        """
        summary = ""
        sys_prompt = str(self.plugin.options['prompt_summarize']['value']) + question
        max_tokens = int(self.plugin.options["summary_max_tokens"]['value'])
        for prompt in chunks:
            self.plugin.window.log("Plugin: web_search:get_summarized_tex [getting summary] "
                                   "(prompt, sys_prompt, max_tokens): {}, {}, {}".
                                   format(prompt, sys_prompt, max_tokens))  # log
            response = self.plugin.window.gpt.quick_call(prompt, sys_prompt, False, max_tokens)
            if response is not None:
                self.plugin.window.log("Plugin: web_search:get_summarized_text [response length]: {}".
                                       format(len(response)))  # log
                summary += response
        return summary

    def get_system_prompt(self, question, prompt):
        """
        Get system prompt from web search results

        :param question: question to search for
        :param prompt: current prompt
        :return: system prompt
        """
        if self.plugin.options["rebuild_question"]['value']:
            use_context = self.plugin.options['rebuild_question_context']['value']
            tmp_prompt = str(self.plugin.options['prompt_question_prefix']['value']) + question
            sys_prompt = str(self.plugin.options['prompt_question']['value']).replace("{time}", time.strftime("%c"))

            self.plugin.window.log("Plugin: web_search:get_system_prompt [getting system prompt] "
                                   "(question, prompt, sys_prompt, tmp_prompt): {}, {}, {}, {}".
                                   format(question, prompt, sys_prompt, tmp_prompt))  # log

            question = self.plugin.window.gpt.quick_call(tmp_prompt, sys_prompt, use_context, 500)
            self.plugin.window.log("Plugin: web_search:get_system_prompt [search question]: {}".format(question))  # log

            if question is not None:
                question = question.replace("<query>", "").replace("</query>", "").strip()
                self.plugin.window.log(
                    "Plugin: web_search:get_system_prompt [search question extracted]: {}".format(question))  # log
                self.plugin.window.log(
                    "Plugin: web_search:get_system_prompt [search question length]: {}".format(len(question)))  # log

        # query the web
        pages = self.query_web(question)

        # get options
        max_per_page = int(self.plugin.options["max_page_content_length"]['value'])
        chunk_size = int(self.plugin.options["chunk_size"]['value'])
        max_prompt_size = int(self.plugin.options["prompt_system_length"]['value'])

        # build chunks from webpages
        chunks = []
        for page in pages:
            if page is None or page == "":
                continue
            if 0 < max_per_page < len(page):
                page = page[:max_per_page]
            chunks += self.to_chunks(page, chunk_size)  # it returns list of chunks

        # log
        self.plugin.window.log(
            "Plugin: web_search:get_system_prompt [num chunks]: {}".format(len(chunks)))  # log

        # build summary
        summary = self.get_summarized_text(chunks, question)

        # log
        self.plugin.window.log(
            "Plugin: web_search:get_system_prompt [summary]: {}".format(summary))  # log
        if summary is not None:
            self.plugin.window.log(
                "Plugin: web_search:get_system_prompt [summary length]: {}".format(len(summary)))  # log

        # append summary to prompt
        if summary is not None and summary.strip() != "":
            prompt += str(self.plugin.options['prompt_system']['value']) + summary.replace("\n", " ").replace("\t", " ")
            if len(prompt) > max_prompt_size:
                prompt = prompt[:max_prompt_size]

        # log
        self.plugin.window.log(
            "Plugin: web_search:get_system_prompt [prompt length]: {}".format(len(prompt)))  # log

        return prompt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.15 02:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction

from ..plugins import Plugins as Handler
from ..utils import trans


class Plugins:
    def __init__(self, window=None):
        """
        Plugins controller

        :param window: main window
        """
        self.window = window
        self.handler = Handler(self.window.config)
        self.config_dialog = False
        self.config_initialized = False
        self.current_plugin = None
        self.enabled = {}

    def toggle_settings(self):
        """Toggles plugins settings dialog"""
        if self.config_dialog:
            self.close_settings()
        else:
            self.open_settings()

    def setup_settings(self):
        """Sets up plugins settings"""
        self.window.plugin_settings.setup()

    def open_settings(self):
        """Opens plugins settings dialog"""
        if not self.config_initialized:
            self.setup_settings()
            self.config_initialized = True
        if not self.config_dialog:
            self.init_settings()
            self.window.ui.dialogs.open('plugin_settings', width=600, height=500)
            self.config_dialog = True

    def init_settings(self):
        """Initializes plugins settings options"""
        selected_plugin = self.current_plugin
        if selected_plugin is None:
            if len(self.handler.plugins) > 0:
                selected_plugin = list(self.handler.plugins.keys())[0]

        for id in self.handler.plugins:
            plugin = self.handler.plugins[id]
            options = plugin.setup()  # get plugin options
            self.current_plugin = id

            for key in options:
                option = options[key]
                option_id = 'plugin.' + id + '.' + key
                if option['type'] == 'text' or option['type'] == 'int' or option['type'] == 'float':
                    if 'slider' in option and option['slider'] \
                            and (option['type'] == 'int' or option['type'] == 'float'):
                        self.config_slider(option_id, option['value'])
                    else:
                        self.config_change(option_id, option['value'])
                elif option['type'] == 'textarea':
                    self.config_change(option_id, option['value'])
                elif option['type'] == 'bool':
                    self.config_toggle(option_id, option['value'])

        self.current_plugin = selected_plugin

    def save_settings(self):
        """Saves plugins settings"""
        selected_plugin = self.current_plugin
        for id in self.handler.plugins:
            plugin = self.handler.plugins[id]
            options = plugin.setup()  # get plugin options

            # add plugin to config if not exists
            if id not in self.window.config.data['plugins']:
                self.window.config.data['plugins'][id] = {}

            self.current_plugin = id
            # update config with new values
            for key in options:
                option = options[key]
                value = None
                if option['type'] == 'int' or option['type'] == 'float':
                    if 'slider' in option and option['slider'] \
                            and (option['type'] == 'int' or option['type'] == 'float'):
                        value = self.window.plugin_option[id][key].slider.value()
                    else:
                        if option['type'] == 'int':
                            try:
                                value = int(self.window.plugin_option[id][key].text())
                            except ValueError:
                                value = 0
                        elif option['type'] == 'float':
                            try:
                                value = float(self.window.plugin_option[id][key].text())
                            except ValueError:
                                value = 0.0
                elif option['type'] == 'text':
                    value = self.window.plugin_option[id][key].text()
                elif option['type'] == 'textarea':
                    value = self.window.plugin_option[id][key].toPlainText()
                elif option['type'] == 'bool':
                    value = self.window.plugin_option[id][key].box.isChecked()
                self.handler.plugins[id].options[key]['value'] = value
                self.window.config.data['plugins'][id][key] = value

            # update config if option not exists
            for key in self.window.config.data['plugins'].keys():
                if key not in self.handler.plugins:
                    self.window.config.data['plugins'].pop(key)

        # save config
        self.window.config.save()
        self.close_settings()
        self.current_plugin = selected_plugin

    def load_defaults(self):
        """Loads plugins settings defaults"""
        self.init_settings()

    def close_settings(self):
        """Closes plugins settings dialog"""
        if self.config_dialog:
            self.window.ui.dialogs.close('plugin_settings')
            self.config_dialog = False

    def register(self, plugin):
        """
        Registers plugin

        :param plugin: Plugin
        """
        self.handler.register(plugin)

    def unregister(self, id):
        """
        Unregisters plugin

        :param id: Plugin id
        """
        if self.handler.is_registered(id):
            self.handler.plugins.pop(id)
            if id in self.enabled:
                self.enabled.pop(id)

    def enable(self, id):
        """
        Enables plugin

        :param id: Plugin id
        """
        if self.handler.is_registered(id):
            self.enabled[id] = True
            self.handler.plugins[id].on_enable()  # call plugin enable method
            self.window.config.data['plugins_enabled'][id] = True
            self.window.config.save()

            # update audio menu
            if id == 'audio_azure':
                self.window.controller.audio.update()

        self.update_info()
        self.update()

    def disable(self, id):
        """
        Disables plugin

        :param id: Plugin id
        """
        if self.handler.is_registered(id):
            self.enabled[id] = False
            self.handler.plugins[id].on_disable()  # call plugin disable method
            self.window.config.data['plugins_enabled'][id] = False
            self.window.config.save()

            # update audio menu
            if id == 'audio_azure':
                self.window.controller.audio.update()

        self.update_info()
        self.update()

    def is_enabled(self, id):
        """
        Checks if plugin is enabled

        :param id: Plugin id
        :return: True if enabled
        """
        if self.handler.is_registered(id):
            if id in self.enabled:
                return self.enabled[id]
        return False

    def toggle(self, id):
        """
        Toggles plugin

        :param id: Plugin id
        """
        if self.handler.is_registered(id):
            if self.is_enabled(id):
                self.disable(id)
            else:
                self.enable(id)

    def update(self):
        """Updates plugins menu"""
        for id in self.window.menu['plugins']:
            self.window.menu['plugins'][id].setChecked(False)

        for id in self.enabled:
            if self.enabled[id]:
                self.window.menu['plugins'][id].setChecked(True)

    def setup_menu(self):
        """Sets up plugins menu"""
        for id in self.handler.plugins:
            if id in self.window.menu['plugins']:
                continue
            default_name = self.handler.plugins[id].name
            trans_key = 'plugin.' + id
            name = trans(trans_key)
            if name == trans_key:
                name = default_name
            self.window.menu['plugins'][id] = QAction(name, self.window, checkable=True)
            self.window.menu['plugins'][id].triggered.connect(
                lambda checked=None, id=id: self.window.controller.plugins.toggle(id))
            self.window.menu['menu.plugins'].addAction(self.window.menu['plugins'][id])

    def set_plugin_by_tab(self, idx):
        """
        Sets current plugin by tab index

        :param idx: tab index
        """
        plugin_idx = 0
        for id in self.handler.plugins:
            if self.handler.plugins[id].options:
                if plugin_idx == idx:
                    self.current_plugin = id
                    break
                plugin_idx += 1

    def setup(self):
        """
        Sets up plugins
        """
        self.setup_menu()
        self.load_config()
        self.update()

    def update_info(self):
        """Updates plugins info"""
        enabled_list = []
        for id in self.handler.plugins:
            if self.is_enabled(id):
                enabled_list.append(self.handler.plugins[id].name)
        tooltip = " + ".join(enabled_list)

        count_str = ""
        c = 0
        if len(self.handler.plugins) > 0:
            for id in self.handler.plugins:
                if self.is_enabled(id):
                    c += 1

        if c > 0:
            count_str = "+ " + str(c) + " " + trans('chatbox.plugins')
        self.window.data['chat.plugins'].setText(count_str)
        self.window.data['chat.plugins'].setToolTip(tooltip)

    def load_config(self):
        """
        Loads plugins config
        """
        for id in self.window.config.data['plugins_enabled']:
            if self.window.config.data['plugins_enabled'][id]:
                self.enable(id)

    def config_toggle(self, id, value):
        """
        Toggles checkbox

        :param id: checkbox option id
        :param value: checkbox option value
        """
        key = id.replace('plugin.' + self.current_plugin + '.', '')
        self.window.plugin_option[self.current_plugin][key].box.setChecked(value)

    def get_option(self, id, key):
        """
        Gets plugin option

        :param id: option id
        :return: option value
        """
        return self.handler.plugins[id].options[key]

    def config_change(self, id, value):
        """
        Changes input value

        :param id: input option id
        :param value: input option value
        """
        key = id.replace('plugin.' + self.current_plugin + '.', '')
        option = self.get_option(self.current_plugin, key)
        if 'type' in option and option['type'] == 'int':
            value = round(int(value), 0)
        elif 'type' in option and option['type'] == 'float':
            value = float(value)

        if 'type' in option and option['type'] == 'int' or option['type'] == 'float':
            if value is not None:
                if 'min' in option and option['min'] is not None and value < option['min']:
                    value = option['min']
                elif 'max' in option and option['max'] is not None and value > option['max']:
                    value = option['max']

        self.window.plugin_option[self.current_plugin][key].setText('{}'.format(value))

    def config_slider(self, id, value, type=None):
        """
        Applies slider + input value

        :param id: option id
        :param value: option value
        :param type: option type (slider, input, None)
        """
        key = id.replace('plugin.' + self.current_plugin + '.', '')
        is_integer = False
        multiplier = 1

        option = self.get_option(self.current_plugin, key)
        if 'type' in option and option['type'] == 'int':
            is_integer = True
        if 'multiplier' in option:
            multiplier = option['multiplier']

        if type != 'slider':
            if is_integer:
                try:
                    value = round(int(value), 0)
                except:
                    value = 0
            else:
                try:
                    value = float(value)
                except:
                    value = 0.0

            if 'min' in option and value < option['min']:
                value = option['min']
            elif 'max' in option and value > option['max']:
                value = option['max']

            self.window.plugin_option[self.current_plugin][key].input.setText(str(value))

        slider_value = round(float(value) * multiplier, 0)
        input_value = value
        if type == 'slider':
            input_value = value / multiplier
            if is_integer:
                input_value = round(int(input_value), 0)

        # update from slider
        if type == 'slider':
            txt = '{}'.format(input_value)
            self.window.plugin_option[self.current_plugin][key].input.setText(txt)

        # update from input
        elif type == 'input':
            if 'min' in option and slider_value < option['min'] * multiplier:
                slider_value = option['min'] * multiplier
            elif 'max' in option and slider_value > option['max'] * multiplier:
                slider_value = option['max'] * multiplier
            self.window.plugin_option[self.current_plugin][key].slider.setValue(slider_value)

        # update from raw value
        else:
            txt = '{}'.format(value)
            self.window.plugin_option[self.current_plugin][key].input.setText(txt)
            self.window.plugin_option[self.current_plugin][key].slider.setValue(slider_value)

    def apply(self, event, data):
        """
        Applies plugins

        :param event: event
        :param data: event data
        """
        for id in self.handler.plugins:
            if self.is_enabled(id):
                data = self.handler.apply(id, event, data)

        return data
